/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

/* Resource definitions for ApacheMonitor.rc and ApacheMonitor.c
 */

#define IDD_DLGSERVICES                 101
#define IDS_APMONITORTITLE              102
#define IDS_APMONITORCLASS              103
#define IDM_RESTORE                     104
#define IDM_EXIT                        105
#define IDI_APSRVMON                    106
#define IDI_ICOSTOP                     107
#define IDI_ICORUN                      108
#define IDC_STATBAR                     109
#define IDC_SSTATUS                     110
#define IDB_BMPSTOP                     111
#define IDB_BMPRUN                      112
#define IDB_BMPHEADER                   114
#define IDL_SERVICES                    115
#define IDL_STDOUT                      116
#define IDC_SSTART                      117
#define IDC_SSTOP                       118
#define IDC_SRESTART                    119
#define IDC_SEXIT                       120
#define IDC_SMANAGER                    121
#define IDD_DLGCONNECT                  122
#define IDC_LREMOTE                     123
#define IDC_LBROWSE                     124
#define IDC_COMPUTER                    125
#define IDC_SCONNECT                    126
#define IDC_SDISCONN                    127
#define IDS_MSG_FIRST                   256
#define IDS_MSG_APPRUNNING              256
#define IDS_MSG_ERROR                   257
#define IDS_MSG_RUNNINGALL              258
#define IDS_MSG_RUNNING                 259
#define IDS_MSG_RUNNINGNONE             260
#define IDS_MSG_NOSERVICES              261
#define IDS_MSG_MNUSERVICES             262
#define IDS_MSG_MNUSHOW                 263
#define IDS_MSG_MNUEXIT                 264
#define IDS_MSG_SRVSTART                265
#define IDS_MSG_SRVSTARTED              266
#define IDS_MSG_SRVSTOP                 267
#define IDS_MSG_SRVSTOPPED              268
#define IDS_MSG_SRVRESTART              269
#define IDS_MSG_SRVRESTARTED            270
#define IDS_MSG_SRVFAILED               271
#define IDS_MSG_SSTART                  272
#define IDS_MSG_SSTOP                   273
#define IDS_MSG_SRESTART                274
#define IDS_MSG_SERVICES                275
#define IDS_MSG_CONNECT                 276
#define IDS_MSG_ECONNECT                277
#define IDS_MSG_LAST                    277
#define IDM_SM_SERVICE                  0x1100
#define IDM_SM_START                    0x1200
#define IDM_SM_STOP                     0x1400
#define IDM_SM_RESTART                  0x1800
#define IDC_STATIC                      -1

